# -*- coding: utf-8 -*-
# text_utils.py
# Common text processing utilities
# https://github.com/evil-mad/plotink
#
# See below for version information
#
# Copyright (c) 2021 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
text_utils.py

Common text processing utilities
https://github.com/evil-mad/plotink

Intended to provide some common interfaces that can be used by
AxiDraw, EggBot, WaterColorBot, and similar machines.
"""

def version():    # Version number for this document
    """Return version number of this script"""
    return "0.2" # Dated 2021-12-14

__version__ = version()


def format_hms (duration, milliseconds=False):
    '''
    Given a number of milliseconds or seconds, return a formatted string
    in the format:
        "12:34:56 (Hours:Minutes:Seconds)" or
        "34:56 (Minutes:Seconds)", or
        "56 Seconds", or
        "5.231 Seconds",
    depending on the duration.
    Times greater than or equal to 10 s are rounded to the nearest second.
    '''
    if milliseconds: # Input units are milliseconds
        duration = duration / 1000.0
    if duration < 10:
        return f'{duration:.3f} Seconds'
    duration_rounded = int(round(duration))
    if duration_rounded < 60:
        return f"{int(round(duration_rounded)):02} Seconds"
    m_elapsed, s_elapsed = divmod(duration_rounded, 60)
    if duration_rounded < 3600:
        out_string =  f"{int(m_elapsed)}:{int(s_elapsed):02}"
        return out_string + " (Minutes, seconds)"
    h_elapsed, m_elapsed = divmod(m_elapsed, 60)
    out_string =  f"{int(h_elapsed)}:{int(m_elapsed):02}:{int(s_elapsed):02}"
    return out_string + " (Hours, minutes, seconds)"


def xml_escape ( input_text ):
    '''
    Replace the five XML special characters with their character entities
    '''
    new_text = input_text.replace('&','&amp;')
    new_text = new_text.replace('<','&lt;')
    new_text = new_text.replace('>','&gt;')
    new_text = new_text.replace('"','&quot;')
    new_text = new_text.replace("'",'&apos;')
    return new_text



# -*- coding: utf-8 -*-
# spatial_grid.py
# part of plotink: https://github.com/evil-mad/plotink
#
# See below for version information
#
# Copyright (c) 2022 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
spatial_grid.py

Specialized flat grid spatial index class for calculating nearest neighbors
"""

import math
from . import plot_utils

def version():    # Version number for this document
    """Return version number of this script"""
    return "1.0" # Dated 2022-01-03

class Index:
    ''' Grid index class '''
    grid = []           # The grid: List of cells, each of which will contain a list of path ends
    adjacents = []      # Adjacency list; list of neighboring cells for each cell
    lookup = []         # List of which grid cell each path end can be found inside.
    path_count = 0      # Initial number of paths
    vertices = None     # The list of start, end vertices for each path
    reverse = False     # Boolean: Are path reversals allowed?
    bin_size_x = 1.0    # Width of grid cells
    bin_size_y = 1.0    # Height of grid cells
    xmin = -1.0         # Grid bounds
    ymin = -1.0         # Grid bounds
    bins_per_side = 3   # Number of bins per side of the grid

    def __init__(self, vertices, bins_per_side, reverse):
        '''
        Given an input list of path vertices and number of bins per side,
        populate a 1D list that represents a linearized 2D grid,
        bins_per_side x bins_per_side in size. Each cell contains a list that
        indicates which path numbers have ends that can be found in that cell.

        Also populate an adjacency list where each cell contains a list of
        which cells are that cell or its neighbors; up to 9 possible.

        And, populate a 1D "reverse" lookup list that gives the grid-cell
        location of each path end.

        Input vertices is a 1D list of elements: [first_vertex, last_vertex]
        for each path. Each vertex is a (x,y) tuple.

        self.bins_per_side is an integer, that once squared gives the number of grid
        cells. Practical minimum of 3, for 9 squares.

        reverse is boolean, indicating whether the paths can be reversed.
        '''

        self.vertices = vertices
        self.reverse = reverse
        self.bins_per_side = bins_per_side
        self.path_count = len(vertices)
        max_bin = bins_per_side - 1 # array index of the largest x or y bin

        self.find_adjacents()

        # Calculate extent of grid:
        self.xmin, self.ymin = math.inf, math.inf
        xmax, ymax = -math.inf, -math.inf

        for [x_1, y_1], [x_2, y_2] in vertices:
            self.xmin = min(self.xmin, x_1)
            xmax = max(xmax, x_1)
            self.ymin = min(self.ymin, y_1)
            ymax = max(ymax, y_1)
            if reverse:
                self.xmin = min(self.xmin, x_2)
                xmax = max(xmax, x_2)
                self.ymin = min(self.ymin, y_2)
                ymax = max(ymax, y_2)

        # Artificially increase size of grid to avoid vertices on the borders:
        shim = (xmax - self.xmin + ymax - self.ymin) / 200
        self.xmin -= shim
        self.ymin -= shim
        xmax += shim
        ymax += shim

        # Calculate bin sizes:
        self.bin_size_x = (xmax - self.xmin) / bins_per_side
        self.bin_size_y = (ymax - self.ymin) / bins_per_side

        # Initialize the "reverse" lookup list:
        if reverse:
            self.lookup = [0 for temp_var in range(2 * self.path_count)]
        else:
            self.lookup = [0 for temp_var in range(self.path_count)]

        # Initialize the grid, with an empty list in each cell:
        self.grid = [[] for index_i in range(self.bins_per_side * self.bins_per_side)]

        for (index_i, [[x_1, y_1], [x_2, y_2]]) in enumerate(vertices):
            x_bin = min(math.floor((x_1 - self.xmin) / self.bin_size_x), max_bin)
            y_bin = min(math.floor((y_1 - self.ymin) / self.bin_size_y), max_bin)
            grid_index = x_bin + self.bins_per_side * y_bin
            self.grid[grid_index].append(index_i)
            self.lookup[index_i] = grid_index # Which grid cell is the path start in?

            if reverse:
                x_bin = min(math.floor((x_2 - self.xmin) / self.bin_size_x), max_bin)
                y_bin = min(math.floor((y_2 - self.ymin) / self.bin_size_y), max_bin)
                grid_index = x_bin + self.bins_per_side * y_bin
                self.grid[grid_index].append(self.path_count + index_i)
                self.lookup[self.path_count + index_i] = grid_index


    def find_adjacents(self):
        '''
        Populate an adjacency list, where each cell contains a list of
        which cells are that cell or its neighbors; up to 9 possible.
        '''
        max_bin = self.bins_per_side - 1

        self.adjacents = [[a_s] for a_s in range(self.bins_per_side * self.bins_per_side)]
        for y_row in range(self.bins_per_side):
            for x_col in range(self.bins_per_side):
                index_i = x_col + y_row * (self.bins_per_side)
                if x_col > 0:
                    self.adjacents[index_i].append(index_i - 1) # OK
                    if y_row > 0:
                        self.adjacents[index_i].append(index_i - self.bins_per_side - 1)
                    if y_row < max_bin:
                        self.adjacents[index_i].append(index_i + self.bins_per_side - 1)
                if x_col < max_bin:
                    self.adjacents[index_i].append(index_i + 1)
                    if y_row > 0:
                        self.adjacents[index_i].append(index_i - self.bins_per_side + 1)
                    if y_row < max_bin:
                        self.adjacents[index_i].append(index_i + self.bins_per_side + 1)
                if y_row > 0:
                    self.adjacents[index_i].append(index_i - self.bins_per_side)
                if y_row < max_bin:
                    self.adjacents[index_i].append(index_i + self.bins_per_side)


    def nearest(self, vertex_in):
        '''
        Find the nearest path end to the given vertex and return its index.
        Input last_vertex is a [x, y] list.

        Method:
            * Locate which grid cell the input vertex is located in.
            * For every vertex in that grid cell, plus the (up to) eight surrounding it,
                check to see if it is the nearest neighbor to the input vertex.
                If so, return the index of that closest vertex.

            * If there are no vertices in those (up to) 9 cells, check the entire rest of
                the document, and find the closest (global) point.

            * If no vertices are found at all, return None

        The neighborhood of up 8 cells surrounding the initial cell serves as a crude
        circular region surrounding the vertex. In most (but not all) cases, the
        nearest point within that region will be the nearest point globally.
        The precision of that "circle" could be improved by using a finer grid, and a
        larger number of adjacent cells to check.
        '''

        max_bin = self.bins_per_side - 1

        # Use max/min to constrain the initial row and column of our first cell to check
        x_bin = max(min(math.floor((vertex_in[0] - self.xmin) / self.bin_size_x), max_bin), 0)
        y_bin = max(min(math.floor((vertex_in[1] - self.ymin) / self.bin_size_y), max_bin), 0)
        last_cell = x_bin + self.bins_per_side * y_bin

        neighborhood_cells = self.adjacents[last_cell].copy()

        best_dist = math.inf
        best_index = None
        for cell in neighborhood_cells:
            for path_index in self.grid[cell]:
                if path_index >= self.path_count: # new path is reversed
                    vertex = self.vertices[path_index - self.path_count][1]
                else:
                    vertex = self.vertices[path_index][0] # Beginning of next path

                dist = plot_utils.square_dist(vertex_in, vertex)
                if dist < best_dist:
                    best_dist = dist
                    best_index = path_index
        if best_index:
            return best_index

        # Fallback: Check remaining cells if no points were found in neighborhood:
        for cell in range (len(self.adjacents)):
            if cell in neighborhood_cells:
                continue
            for path_index in self.grid[cell]:
                if path_index >= self.path_count: # new path is reversed
                    vertex = self.vertices[path_index - self.path_count][1]
                else:
                    vertex = self.vertices[path_index][0]

                dist = plot_utils.square_dist(vertex_in, vertex)
                if dist < best_dist:
                    best_dist = dist
                    best_index = path_index
        return best_index


    def remove_path(self, path_index):
        '''
        Remove the vertex with the given path_index from the spatial index.
        Input path_index must be < self.path_count.

        If reversing is enabled, also remove the vertex with index
        path_index + self.path_count
        '''

        cell_number = self.lookup[path_index]
        self.grid[cell_number].remove(path_index)

        if self.reverse:
            other_index = path_index + self.path_count
            cell_number = self.lookup[other_index]
            self.grid[cell_number].remove(other_index)



from importlib import import_module
import sys
import os

'''
this handles importing in two major cases

In one case, the program is being run as a python script/library without inkscape. Everything can proceed normally.

In the other, AxiDraw's dependencies are siloed into DEPENDENCY_DIR_NAME so as not to mess with other installations/extensions/etc, and to import them, the import system has to be instructed to look in DEPENDENCY_DIR_NAME first.

In packages that need to use `from_dependency_import` a file named `plot_utils_import.py` containing `from plotink.plot_utils_import import * has to be there, so that this importing mechanism is available to files in the package with `from plot_utils_import import from_dependency_import`

# to think about this might be a good place to put a blanket import lxml
'''

DEPENDENCY_DIR_NAME = 'axidraw_deps'
DEPENDENCY_DIR = os.path.join(os.path.abspath(os.getcwd()), DEPENDENCY_DIR_NAME)
INK_EXTENSIONS_DIR = os.path.join(DEPENDENCY_DIR, 'ink_extensions')

def from_dependency_import(module_name):
    ''' module_name ex: "ink_extensions", "ink_extensions.inkex"
    module_name must be the name of a module, not a class, function, etc. '''
    module = None

    if os.path.isdir(DEPENDENCY_DIR):
        # running as an inkscape extension in inkscape

        sys.path.insert(0, DEPENDENCY_DIR)

        # inkscape-provided files don't know they are
        # in the ink_extensions module
        sys.path.insert(0, INK_EXTENSIONS_DIR)

        try:
            module = import_module(module_name)
        finally:
            for folder in [DEPENDENCY_DIR, INK_EXTENSIONS_DIR]:
                if folder in sys.path:
                    sys.path.remove(folder)

    else:
        # running as a python module with traditionally installed packages
        # e.g. if you used pip or setup.py
        module = import_module(module_name)

    return module



# -*- coding: utf-8 -*-
# plot_utils.py
# Common plotting utilities
# https://github.com/evil-mad/plotink
#
# See below for version information
#
# Copyright (c) 2021 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
plot_utils.py

Common plotting utilities
https://github.com/evil-mad/plotink

Intended to provide some common interfaces that can be used by
EggBot, WaterColorBot, AxiDraw, and similar machines.
"""

from math import sqrt, isclose

from .plot_utils_import import from_dependency_import
cspsubdiv = from_dependency_import('ink_extensions.cspsubdiv')
simplepath = from_dependency_import('ink_extensions.simplepath')
bezmisc = from_dependency_import('ink_extensions.bezmisc')
ffgeom = from_dependency_import('ink_extensions.ffgeom')

def version():    # Version number for this document
    """Return version number of this script"""
    return "0.3" # Dated 2021-12-26

__version__ = version()

PX_PER_INCH = 96.0
# This value has changed to 96 px per inch, as of version 0.12 of this library.
# Prior versions used 90 PPI, corresponding the value used in Inkscape < 0.92.
# For use with Inkscape 0.91 (or older), use PX_PER_INCH = 90.0

trivial_svg = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <svg
       xmlns:dc="http://purl.org/dc/elements/1.1/"
       xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
       xmlns:svg="http://www.w3.org/2000/svg"
       xmlns="http://www.w3.org/2000/svg"
       version="1.1"
       id="svg15158"
       viewBox="0 0 297 210"
       height="210mm"
       width="297mm">
    </svg>
    """

def checkLimits(value, lower_bound, upper_bound):
    """
    Limit a value to within a range.
    Return constrained value with error boolean.
    """
    if value > upper_bound:
        return upper_bound, True
    if value < lower_bound:
        return lower_bound, True
    return value, False


def checkLimitsTol(value, lower_bound, upper_bound, tolerance):
    """
    Limit a value to within a range.
    Return constrained value with error boolean.
    Allow a range of tolerance where we constrain the value without an error message.
    """
    if value > upper_bound:
        if value > (upper_bound + tolerance):
            return upper_bound, True  # Truncate & throw error
        return upper_bound, False  # Truncate with no error
    if value < lower_bound:
        if value < (lower_bound - tolerance):
            return lower_bound, True  # Truncate & throw error
        return lower_bound, False  # Truncate with no error
    return value, False  # Return original value without error


def point_in_bounds(point, bounds, tolerance=1e-9):
    """
    Given an input point [x, y], a rectangular bounding region
    [[x_min, y_min], [x_max, y_max]], and an optional tolerance,
    return True if the point is within the bounds, to within the
    required tolerance, and False otherwise.
    """
    x, y = point
    [[x_min, y_min], [x_max, y_max]] = bounds
    if x < x_min - tolerance:
        return False
    if y < y_min - tolerance:
        return False
    if x > x_max + tolerance:
        return False
    if y > y_max + tolerance:
        return False
    return True


def clip_code(x_in, y_in, x_min, x_max, y_min, y_max):
    """Encode point position with respect to boundary box"""
    code = 0
    if x_in < x_min:
        code = 1  # Left
    if x_in > x_max:
        code |= 2 # Right
    if y_in < y_min:
        code |= 4 # Top
    if y_in > y_max:
        code |= 8 # Bottom
    return code


def clip_segment(segment, bounds):
    """
    Given an input line segment [[x_1, y_1], [x_2, y_2]], as well as a
    rectangular bounding region [[x_min, y_min], [x_max, y_max]], clip and
    keep the part of the segment within the bounding region, using the
    Cohen–Sutherland algorithm.
    Return a boolean value, "accept", indicating that the output
    segment is non-empty, as well as truncated segment,
    [[x_1', y_1'], [x_2', y_2']], giving the portion of the input line segment
    that fits within the bounds.
    """

    x_1 = segment[0][0]
    y_1 = segment[0][1]
    x_2 = segment[1][0]
    y_2 = segment[1][1]

    x_min = bounds[0][0]
    y_min = bounds[0][1]
    x_max = bounds[1][0]
    y_max = bounds[1][1]

    iterations = 0

    while True: # Repeat until return
        code_1 = clip_code(x_1, y_1, x_min, x_max, y_min, y_max)
        code_2 = clip_code(x_2, y_2, x_min, x_max, y_min, y_max)

        # Trivial accept:
        if code_1 == 0 and code_2 == 0:
            return True, segment # Both endpoints are within bounds.
        # Trivial reject, if both endpoints are outside, and on the same side:
        if code_1 & code_2:
            return False, segment # Verify with bitwise AND.
        if iterations > 3: # Failsafe; exit if the value has not converged;
            return True, segment # Avoids infinite loops near precision limits.

        # Otherwise, at least one point is out of bounds; not trivial.
        if code_1 != 0:
            code = code_1
        else:
            code = code_2

        # Clip at a single boundary; may need to do this up to twice per vertex

        if code & 1: # Vertex on LEFT side of bounds:
            x_new = x_min  # Find intersection of our segment with x_min
            slope = (y_2 - y_1) / (x_2 - x_1)
            y_new = slope * (x_min - x_1) + y_1

        elif code & 2:  # Vertex on RIGHT side of bounds:
            x_new = x_max # Find intersection of our segment with x_max
            slope = (y_2 - y_1) / (x_2 - x_1)
            y_new = slope * (x_max - x_1) + y_1

        elif code & 4: # Vertex on TOP side of bounds:
            y_new = y_min  # Find intersection of our segment with y_min
            slope = (x_2 - x_1) / (y_2 - y_1)
            x_new = slope * (y_min - y_1) + x_1

        elif code & 8: # Vertex on BOTTOM side of bounds:
            y_new = y_max  # Find intersection of our segment with y_max
            slope = (x_2 - x_1) / (y_2 - y_1)
            x_new = slope * (y_max - y_1) + x_1

        if code == code_1:
            x_1 = x_new
            y_1 = y_new
        else:
            x_2 = x_new
            y_2 = y_new
        segment = [[x_1, y_1], [x_2, y_2]] # Now checking this clipped segment
        iterations += 1


def constrainLimits(value, lower_bound, upper_bound):
    """ Limit a value to within a range. """
    return max(lower_bound, min(upper_bound, value))


def distance(x_in, y_in):
    """
    Pythagorean theorem
    """
    return sqrt(x_in * x_in + y_in * y_in)


def dotProductXY(input_vector_first, input_vector_second):
    """Dot product of vectors"""
    temp = input_vector_first[0] * input_vector_second[0] +\
                    input_vector_first[1] * input_vector_second[1]
    if temp > 1:
        return 1
    if temp < -1:
        return -1
    return temp


def getLength(altself, name, default):
    """
    Get the <svg> attribute with name "name" and default value "default"
    Parse the attribute into a value and associated units.  Then, accept
    no units (''), units of pixels ('px'), and units of percentage ('%').
    Return value in px.
    """
    string_to_parse = altself.document.getroot().get(name)

    if string_to_parse:
        value, unit = parseLengthWithUnits(string_to_parse)
        if value is None:
            return None
        if unit in ('', 'px'):
            return float(value)
        if unit == 'in':
            return float(value) * PX_PER_INCH
        if unit == 'mm':
            return float(value) * PX_PER_INCH / 25.4
        if unit == 'cm':
            return float(value) * PX_PER_INCH / 2.54
        if unit in ('Q', 'q'):
            return float(value) * PX_PER_INCH / (40.0 * 2.54)
        if unit == 'pc':
            return float(value) * PX_PER_INCH / 6.0
        if unit == 'pt':
            return float(value) * PX_PER_INCH / 72.0
        if unit == '%':
            return float(default) * value / 100.0
        # Unsupported units
        return None
    # No width specified; assume the default value
    return float(default)


def getLengthInches(altself, name):
    """
    Get the <svg> attribute with name "name", and parse it as a length,
    into a value and associated units. Return value in inches.

    As of version 0.11, units of 'px' or no units ('') are interpreted
    as imported px, at a resolution of 96 px per inch, as per the SVG
    specification. (Prior versions returned None in this case.)

    This allows certain imported SVG files, (imported with units of px)
    to plot while they would not previously. However, it may also cause
    new scaling issues in some circumstances. Note, for example, that
    Adobe Illustrator uses 72 px per inch, and Inkscape used 90 px per
    inch prior to version 0.92.
    """
    string_to_parse = altself.document.getroot().get(name)
    if string_to_parse:
        value, unit = parseLengthWithUnits(string_to_parse)
        if value is None:
            return None
        if unit == 'in':
            return float(value)
        if unit == 'mm':
            return float(value) / 25.4
        if unit == 'cm':
            return float(value) / 2.54
        if unit in ('Q', 'q'):
            return float(value) / (40.0 * 2.54)
        if unit == 'pc':
            return float(value) / 6.0
        if unit == 'pt':
            return float(value) / 72.0
        if unit in ('', 'px'):
            return float(value) / 96.0
    # Unsupported units (including '%') or no string to parse:
    return None


def parseLengthWithUnits(string_to_parse):
    """
    Parse an SVG value which may or may not have units attached.
    There is a more general routine to consider in scour.py if more
    generality is ever needed.
    """

    if string_to_parse is None:
        return None, None
    units = 'px'
    string = string_to_parse.strip()
    if string[-2:] == 'px':  # pixels, at a size of PX_PER_INCH per inch
        string = string[:-2]
    elif string[-2:] == 'in':  # inches
        string = string[:-2]
        units = 'in'
    elif string[-2:] == 'mm':  # millimeters
        string = string[:-2]
        units = 'mm'
    elif string[-2:] == 'cm':  # centimeters
        string = string[:-2]
        units = 'cm'
    elif string[-2:] == 'pt':  # points; 1pt = 1/72th of 1in
        string = string[:-2]
        units = 'pt'
    elif string[-2:] == 'pc':  # picas; 1pc = 1/6th of 1in
        string = string[:-2]
        units = 'pc'
    elif string[-1:] == 'Q' or string[-1:] == 'q':  # quarter-millimeters. 1q = 1/40th of 1cm
        string = string[:-1]
        units = 'Q'
    elif string[-1:] == '%':
        units = '%'
        string = string[:-1]

    try:
        value = float(string)
    except ValueError:
        return None, None

    return value, units


def unitsToUserUnits(input_string, percent_ref=None):
    """
    Custom replacement for the unittouu routine in inkex.py

    Parse input_string into a value and units. Return value in user units (typically "px").
    If input_string units are '%', then return (value * percent_ref / 100), where percent_ref
    is a number corresponding to a length of 100% in in user units, e.g., document width.
    """
    value, unit = parseLengthWithUnits(input_string)
    if value is None:
        return None
    if unit in ('', 'px'):
        return float(value)
    if unit == 'in':
        return float(value) * PX_PER_INCH
    if unit == 'mm':
        return float(value) * PX_PER_INCH / 25.4
    if unit == 'cm':
        return float(value) * PX_PER_INCH / 2.54
    if unit in ('Q', 'q'):
        return float(value) * PX_PER_INCH / 101.6
    if unit == 'pc':
        return float(value) * PX_PER_INCH / 6.0
    if unit == 'pt':
        return float(value) * PX_PER_INCH / 72.0
    if unit == '%':
        if percent_ref:
            return float(value) * float(percent_ref) / 100.0
        else:
            return float(value) / 100.0
    return None # Handle case of cnsupported units


def position_scale (x_value, y_value, units_code):
    '''
    Format XY position data to be returned to user
    x_value, y_value inputs are in inches.
    Output set by units_code: 1 for cm, 2 for mm, 0 (or otherwise) for inch.
    '''
    if units_code == 1 : # If using centimeter units
        x_value = x_value * 2.54
        y_value = y_value * 2.54
    if units_code == 2: # If using millimeter units
        x_value = x_value * 25.4
        y_value = y_value * 25.4
    return x_value, y_value


def subdivideCubicPath(s_p, flat, i=1):
    """
    Break up a bezier curve into smaller curves, each of which
    is approximately a straight line within a given tolerance
    (the "smoothness" defined by [flat]).

    This is a modified version of cspsubdiv.cspsubdiv(). I rewrote the recursive
    call because it caused recursion-depth errors on complicated line segments.
    """

    while True:
        while True:
            if i >= len(s_p):
                return
            p_0 = s_p[i - 1][1]
            p_1 = s_p[i - 1][2]
            p_2 = s_p[i][0]
            p_3 = s_p[i][1]

            b_list = (p_0, p_1, p_2, p_3)

            if cspsubdiv.maxdist(b_list) > flat:
                break
            i += 1

        one, two = bezmisc.beziersplitatt(b_list, 0.5)
        s_p[i - 1][2] = one[1]
        s_p[i][0] = two[2]
        p_list = [one[2], one[3], two[1]]
        s_p[i:1] = [p_list]


def max_dist_from_n_points(input_points):
    """
    Like cspsubdiv.maxdist, but it can check for distances of any number of points >= 0.

    `input_points` is an ordered collection of xy vertices.
    The first point and the last point define the segment we are finding distances from.

    does not mutate `input_points`
    """
    assert len(input_points) >= 3, "There must be points (other than begin/end) to check."

    points = [ffgeom.Point(point[0], point[1]) for point in input_points]
    segment = ffgeom.Segment(points.pop(0), points.pop())

    distances = [segment.distanceToPoint(point) for point in points]
    return max(distances)


def supersample(vertices, tolerance):
    """
    Given a list of vertices, remove some according to the following algorithm.

    Suppose that the vertex list consists of points A, B, C, D, E, and so forth,
    which define segments AB, BC, CD, DE, EF, and so on.

    We first test to see if vertex B can be removed, by using perpDistanceToPoint
    to check whether the distance between B and segment AC is less than tolerance.

    If B can be removed, then check to see if the next vertex, C, can be removed.
    Both B and C can be removed if the both the distance between B and AD is less
    than Tolerance and the distance between C and AD is less than Tolerance. Continue
    removing additional vertices, so long as the perpendicular distance between every
    point removed and the resulting segment is less than tolerance (and the end of the
    vertex list is not reached).

    If B cannot be removed, then move onto vertex C, and perform the same checks,
    until the end of the vertex list is reached.
    """
    if len(vertices) <= 2: # there is nothing to delete
        return vertices

    start_index = 0 # can't remove first vertex
    while start_index < len(vertices) - 2:
        end_index = start_index + 2
        # test the removal of (start_index, end_index), exclusive until we can't advance end_index
        while (max_dist_from_n_points(vertices[start_index:end_index + 1]) < tolerance
               and end_index < len(vertices)):
            end_index += 1 # try removing the next vertex too

        vertices[start_index + 1:end_index - 1] = [] # delete (start_index, end_index), exclusive
        start_index += 1

def userUnitToUnits(distance_uu, unit_string):
    """
    Custom replacement for the uutounit routine in inkex.py

    Parse the attribute into a value and associated units.
    Return value in user units (typically "px").
    """

    if distance_uu is None:  # Couldn't parse the value
        return None
    if unit_string in ('', 'px'):
        return float(distance_uu)
    if unit_string == 'in':
        return float(distance_uu) / PX_PER_INCH
    if unit_string == 'mm':
        return float(distance_uu) / (PX_PER_INCH / 25.4)
    if unit_string == 'cm':
        return float(distance_uu) / (PX_PER_INCH / 2.54)
    if unit_string in ('Q', 'q'):
        return float(distance_uu) / (PX_PER_INCH / (40.0 * 2.54))
    if unit_string == 'pc':
        return float(distance_uu) / (PX_PER_INCH / 6.0)
    if unit_string == 'pt':
        return float(distance_uu) / (PX_PER_INCH / 72.0)
    if unit_string == '%':
        return float(distance_uu) * 100.0
    # Unsupported units
    return None


def vb_scale(v_b, p_a_r, doc_width, doc_height):
    """"
    Parse SVG viewbox and generate scaling parameters.
    Reference documentation: https://www.w3.org/TR/SVG11/coords.html

    Inputs:
        v_b:         Contents of SVG viewbox attribute
        p_a_r:      Contents of SVG preserveAspectRatio attribute
        doc_width:  Width of SVG document
        doc_height: Height of SVG document

    Output: s_x, s_y, o_x, o_y
        Scale parameters (s_x,s_y) and offset parameters (o_x,o_y)

    """
    if v_b is None:
        return 1, 1, 0, 0 # No viewbox; return default transform
    vb_array = v_b.strip().replace(',', ' ').split()

    if len(vb_array) < 4:
        return 1, 1, 0, 0 # invalid viewbox; return default transform

    min_x = float(vb_array[0]) # viewbox offset: x
    min_y = float(vb_array[1]) # viewbox offset: y
    width = float(vb_array[2]) # viewbox width
    height = float(vb_array[3]) # viewbox height

    if width <= 0 or height <= 0:
        return 1, 1, 0, 0 # invalid viewbox; return default transform

    d_width = float(doc_width)
    d_height = float(doc_height)

    if d_width <= 0 or d_height <= 0:
        return 1, 1, 0, 0 # invalid document size; return default transform

    ar_doc = d_height / d_width # Document aspect ratio
    ar_vb = height / width      # viewbox aspect ratio

    # Default values of the two preserveAspectRatio parameters:
    par_align = "xmidymid" # "align" parameter (lowercased)
    par_mos = "meet"       # "meetOrSlice" parameter

    if p_a_r is not None:
        par_array = p_a_r.strip().replace(',', ' ').lower().split()
        if len(par_array) > 0:
            par0 = par_array[0]
            if par0 == "defer":
                if len(par_array) > 1:
                    par_align = par_array[1]
                    if len(par_array) > 2:
                        par_mos = par_array[2]
            else:
                par_align = par0
                if len(par_array) > 1:
                    par_mos = par_array[1]

    if par_align == "none":
        # Scale document to fill page. Do not preserve aspect ratio.
        # This is not default behavior, nor what happens if par_align
        # is not given; the "none" value must be _explicitly_ specified.

        s_x = d_width/ width
        s_y = d_height / height
        o_x = -min_x
        o_y = -min_y
        return s_x, s_y, o_x, o_y

    # Other than "none", all situations fall into two classes:
    #
    # 1)   (ar_doc >= ar_vb AND par_mos == "meet")
    #        or  (ar_doc < ar_vb AND par_mos == "slice")
    #     -> In these cases, scale document up until viewbox fills doc in X.
    #
    # 2)   All other cases, i.e.,
    #     (ar_doc < ar_vb AND par_mos == "meet")
    #        or  (ar_doc >= ar_vb AND par_mos == "slice")
    #     -> In these cases, scale document up until viewbox fills doc in Y.
    #
    # Note in cases where the scaled viewbox exceeds the document
    # (page) boundaries (all "slice" cases and many "meet" cases where
    # an offset value is given) that this routine does not perform
    # any clipping, but subsequent clipping to the page boundary
    # is appropriate.
    #
    # Besides "none", there are 9 possible values of par_align:
    #     xminymin xmidymin xmaxymin
    #     xminymid xmidymid xmaxymid
    #     xminymax xmidymax xmaxymax

    if (((ar_doc >= ar_vb) and (par_mos == "meet"))
            or ((ar_doc < ar_vb) and (par_mos == "slice"))):
        # Case 1: Scale document up until viewbox fills doc in X.

        s_x = d_width / width
        s_y = s_x # Uniform aspect ratio
        o_x = -min_x

        scaled_vb_height = ar_doc * width
        excess_height = scaled_vb_height - height

        if par_align in {"xminymin", "xmidymin", "xmaxymin"}:
            # Case: Y-Min: Align viewbox to minimum Y of the viewport.
            o_y = -min_y

        elif par_align in {"xminymax", "xmidymax", "xmaxymax"}:
            # Case: Y-Max: Align viewbox to maximum Y of the viewport.
            o_y = -min_y + excess_height

        else: # par_align in {"xminymid", "xmidymid", "xmaxymid"}:
            # Default case: Y-Mid: Center viewbox on page in Y
            o_y = -min_y + excess_height / 2

        return s_x, s_y, o_x, o_y

    # Case 2: Scale document up until viewbox fills doc in Y.

    s_y = d_height / height
    s_x = s_y # Uniform aspect ratio
    o_y = -min_y

    scaled_vb_width = height / ar_doc
    excess_width = scaled_vb_width - width

    if par_align in {"xminymin", "xminymid", "xminymax"}:
        # Case: X-Min: Align viewbox to minimum X of the viewport.
        o_x = -min_x

    elif par_align in {"xmaxymin", "xmaxymid", "xmaxymax"}:
        # Case: X-Max: Align viewbox to maximum X of the viewport.
        o_x = -min_x + excess_width

    else: # par_align in {"xmidymin", "xmidymid", "xmidymax"}:
        # Default case: X-Mid: Center viewbox on page in X
        o_x = -min_x + excess_width / 2

    return s_x, s_y, o_x, o_y
    # return 1, 1, 0, 0 # Catch-all: return default transform


def points_equal(point_a, point_b):
    """
    Given two vertices point_a and point_b, each a 2-tuple,
    determine if the two points are close enough to be considered "equal"
    with a floating-point-friendly "fuzzy" comparison.
    """
    return isclose(point_a[0], point_b[0]) and isclose(point_a[1], point_b[1])


def points_near(point_a, point_b, squared_tolerance):
    """
    Given two vertices point_a and point_b, each a 2-tuple, return True if the two
    points are coincident to within a certain tolerance.

    Arguments:
        point_a, point_b:  Vertex (x,y), 2-tuples of floats
        squared_tolerance: Square of maximum allowed distance between vertices

    if (point_a.x - point_b.x)^2 + (point_a.y - point_b.y)^2 < tolerance^2,
        then return True.
    """
    delta_x = point_a[0] - point_b[0]
    delta_y = point_a[1] - point_b[1]

    return (delta_x * delta_x + delta_y * delta_y) < squared_tolerance


def square_dist(point_a, point_b):
    """
    Given two vertices point_a and point_b, each a 2-tuple,
    return the square of the distance between them.
    """
    delta_x = point_a[0] - point_b[0]
    delta_y = point_a[1] - point_b[1]
    return delta_x * delta_x + delta_y * delta_y


def vInitial_VF_A_Dx(v_final, acceleration, delta_x):
    """
    Kinematic calculation: Maximum allowed initial velocity to arrive at distance X
    with specified final velocity, and given maximum linear acceleration.

    Calculate and return the (real) initial velocity, given an final velocity,
        acceleration rate, and distance interval.

    Uses the kinematic equation Vi^2 = Vf^2 - 2 a D_x , where
            Vf is the final velocity,
            a is the acceleration rate,
            D_x (delta x) is the distance interval, and
            Vi is the initial velocity.

    We are looking at the positive root only-- if the argument of the sqrt
        is less than zero, return -1, to indicate a failure.
    """
    initial_v_squared = (v_final * v_final) - (2 * acceleration * delta_x)
    if initial_v_squared >= 0:
        return sqrt(initial_v_squared)
    return -1


def vFinal_Vi_A_Dx(v_initial, acceleration, delta_x):
    """
    Kinematic calculation: Final velocity with constant linear acceleration.

    Calculate and return the (real) final velocity, given an initial velocity,
        acceleration rate, and distance interval.

    Uses the kinematic equation Vf^2 = 2 a D_x + Vi^2, where
            Vf is the final velocity,
            a is the acceleration rate,
            D_x (delta x) is the distance interval, and
            Vi is the initial velocity.

    We are looking at the positive root only-- if the argument of the sqrt
        is less than zero, return -1, to indicate a failure.
    """
    final_v_squared = (2 * acceleration * delta_x) + (v_initial * v_initial)
    if final_v_squared >= 0:
        return sqrt(final_v_squared)
    return -1


def pathdata_first_point(path):
    """
    Return the first (X,Y) point from an SVG path data string

    Input:  A path data string; the text of the 'd' attribute of an SVG path
    Output: Two floats in a list representing the x and y coordinates of the first point
    """

    parsed_path = simplepath.parsePath(path) # parsePath splits path into segments

    for command, params in parsed_path:
        if command == 'M':
            return [params[0], params[1]]

    # Properly constructed paths begin with a moveto ('M') command.
    return None


def pathdata_last_point(path):
    """
    Return the last (X,Y) point from an SVG path data string

    Input:  A path data string; the text of the 'd' attribute of an SVG path
    Output: Two floats in a list representing the x and y coordinates of the last point
    """
    parsed_path = simplepath.parsePath(path) # parsePath splits path into segments
    command, params = parsed_path[-1] # look at the last command to determine the last point

    if command.upper() == 'Z':
        # find the last move command, since Z moves to the start of the last subpath
        for command, params in reversed(parsed_path[:-1]):
            if command == 'M':
                return [params[0], params[1]]

        # paths must have at least one moveto, so we should never get here.
        return None

    # Otherwise: The last command should be in the set 'MLCQA'
    #     - All commands converted to absolute by parsePath.
    #     - Can ignore Z (case handled)
    #     - Can ignore H,V, since those are converted to L by parsePath.
    #     - Can ignore S, converted to C by parsePath.
    #     - Can ignore T, converted to Q by parsePath.
    #
    #     MLCQA: Commands all ending in (X,Y) pair.

    x_val = params[-2] # Second to last parameter given
    y_val = params[-1] # Last parameter given

    return [x_val, y_val]



# coding=utf-8
# ebb_serial.py
# Serial connection utilities for EiBotBoard
# https://github.com/evil-mad/plotink
#
# Intended to provide some common interfaces that can be used by
# EggBot, WaterColorBot, AxiDraw, and similar machines.
#
# See below for version information
#
# Thanks to Shel Michaels for bug fixes and helpful suggestions.
#
# The MIT License (MIT)
#
# Copyright (c) 2021 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from distutils.version import LooseVersion
import gettext
import logging

from .plot_utils_import import from_dependency_import
inkex = from_dependency_import('ink_extensions.inkex')
serial = from_dependency_import('serial')

logger = logging.getLogger(__name__)

def version():      # Version number for this document
    return "0.16"   # Dated 2021-10-02


def findPort():
    # Find first available EiBotBoard by searching USB ports.
    # Return serial port object.
    try:
        from serial.tools.list_ports import comports
    except ImportError:
        return None
    if comports:
        try:
            com_ports_list = list(comports())
        except TypeError: # https://github.com/evil-mad/plotink/issues/38
            return None
        ebb_port = None
        for port in com_ports_list:
            if port[1].startswith("EiBotBoard"):
                ebb_port = port[0]  # Success; EBB found by name match.
                break  # stop searching-- we are done.
        if ebb_port is None:
            for port in com_ports_list:
                if port[2].startswith("USB VID:PID=04D8:FD92"):
                    ebb_port = port[0]  # Success; EBB found by VID/PID match.
                    break  # stop searching-- we are done.
        return ebb_port


def find_named_ebb(port_name):
    # Find a specific EiBotBoard identified by a string giving either:
    #        The enumerated serial port, or
    #        An EBB "Name tag"
    #
    # Names should be 3-16 characters long.
    # Comparisons are not case sensitive.
    #
    # (Name tags may assigned with the ST command on firmware 2.5.5 and later.)
    #
    # If found:     Return serial port name (enumeration)
    # If not found, Return None
    if port_name is not None:
        try:
            from serial.tools.list_ports import comports
        except ImportError:
            return None
        if comports:
            needle = 'SER=' + port_name     # pyserial 3
            needle2 = 'SNR=' + port_name    # pyserial 2.7
            needle3 = '(' + port_name + ')' # e.g., "(COM4)"

            needle = needle.lower()
            needle2 = needle2.lower()
            needle3 = needle3.lower()
            plower = port_name.lower()

            com_ports_list = list(comports())
            ebb_port = None

            for port in com_ports_list:
                p0 = port[0].lower()
                p1 = port[1].lower()
                p2 = port[2].lower()

                if needle in p2:
                    return port[0]  # Success; EBB found by name match.
                if needle2 in p2:
                    return port[0]  # Success; EBB found by name match.
                if needle3 in p1:
                    return port[0]  # Success; EBB found by port match.

                p1 = p1[11:]
                if p1.startswith(plower):
                    return port[0]  # Success; EBB found by name match.
                if p0.startswith(plower):
                    return port[0]  # Success; EBB found by port match.

                needle.replace(" ", "_") # SN on Windows has underscores, not spaces.
                if needle in p2:
                    return port[0]  # Success; EBB found by port match.

                needle2.replace(" ", "_") # SN on Windows has underscores, not spaces.
                if needle2 in p2:
                    return port[0]  # Success; EBB found by port match.


def query_nickname(port_name, verbose=True):
    # Query the EBB nickname and report it.
    # This feature is only supported in firmware versions 2.5.5 and newer.
    # If verbose is True or omitted, the result will be human readable.
    # A short version is returned if verbose is False.
    # http://evil-mad.github.io/EggBot/ebb.html#QT
    if port_name is not None:
        version_status = min_version(port_name, "2.5.5")

        if version_status:
            raw_string = (query(port_name, 'QT\r'))
            if raw_string.isspace():
                if verbose:
                    return "This AxiDraw does not have a nickname assigned."
                else:
                    return None
            else:
                if verbose:
                    return "AxiDraw nickname: " + raw_string
                else:
#                     string_temp = str(raw_string).strip()
                    return str(raw_string).strip()
        elif version_status is False:
            if verbose:
                return "AxiDraw naming requires firmware version 2.5.5 or higher."

def write_nickname(port_name, nickname):
    # Write the EBB nickname.
    # This feature is only supported in firmware versions 2.5.5 and newer.
    # http://evil-mad.github.io/EggBot/ebb.html#ST
    if port_name is not None:
        version_status = min_version(port_name, "2.5.5")

        if version_status:
            try:
                cmd = 'ST,' + nickname + '\r'
                command(port_name,cmd)
                return True
            except:
                return False

def reboot(port_name):
    # Reboot the EBB, as though it were just powered on.
    # This feature is only supported in firmware versions 2.5.5 and newer.
    # It has no effect if called on an EBB with older firmware.
    # http://evil-mad.github.io/EggBot/ebb.html#RB
    if port_name is not None:
        version_status = min_version(port_name, "2.5.5")
        if version_status:
            try:
                command(port_name,'RB\r')
            except:
                pass


def list_port_info():
    # Find and return a list of all USB devices and their information.
    try:
        from serial.tools.list_ports import comports
    except ImportError:
        return None
    if comports:
        com_ports_list = list(comports())
        port_info_list = []
        for port in com_ports_list:
            port_info_list.append(port[0]) # port name
            port_info_list.append(port[1]) # Identifier
            port_info_list.append(port[2]) # VID/PID
        if port_info_list:
            return port_info_list


def listEBBports():
    # Find and return a list of all EiBotBoard units
    # connected via USB port.
    try:
        from serial.tools.list_ports import comports
    except ImportError:
        return None
    if comports:
        com_ports_list = list(comports())
        ebb_ports_list = []
        for port in com_ports_list:
            port_has_ebb = False
            if port[1].startswith("EiBotBoard"):
                port_has_ebb = True
            elif port[2].startswith("USB VID:PID=04D8:FD92"):
                port_has_ebb = True
            if port_has_ebb:
                ebb_ports_list.append(port)
        if ebb_ports_list:
            return ebb_ports_list


def list_named_ebbs():
    # Return discriptive list of all EiBotBoard units
    ebb_ports_list = listEBBports()
    if not ebb_ports_list:
        return
    ebb_names_list = []
    for port in ebb_ports_list:
        name_found = False
        p0 = port[0]
        p1 = port[1]
        p2 = port[2]
        if p1.startswith("EiBotBoard"):
            temp_string = p1[11:]
            if (temp_string):
                if temp_string is not None:
                    ebb_names_list.append(temp_string)
                    name_found = True
        if not name_found:
            # Look for "SER=XXXX LOCAT" pattern,
            #  typical of Pyserial 3 on Windows.
            if 'SER=' in p2 and ' LOCAT' in p2:
                index1 = p2.find('SER=') + len('SER=')
                index2 = p2.find(' LOCAT', index1)
                temp_string = p2[index1:index2]
                if len(temp_string) < 3:
                    temp_string = None
                if temp_string is not None:
                    ebb_names_list.append(temp_string)
                    name_found = True
        if not name_found:
            # Look for "...SNR=XXXX" pattern,
            #  typical of Pyserial 2.7 on Windows,
            #  as in Inkscape 0.91 on Windows
            if 'SNR=' in p2:
                index1 = p2.find('SNR=') + len('SNR=')
                index2 = len(p2)
                temp_string = p2[index1:index2]
                if len(temp_string) < 3:
                    temp_string = None
                if temp_string is not None:
                    ebb_names_list.append(temp_string)
                    name_found = True
        if not name_found:
            ebb_names_list.append(p0)
    return ebb_names_list


def testPort(port_name):
    """
    Open a given serial port, verify that it is an EiBotBoard,
    and return a SerialPort object that we can reference later.

    This routine only opens the port;
    it will need to be closed as well, for example with closePort( port_name ).
    You, who open the port, are responsible for closing it as well.

    """
    if port_name is not None:
        try:
            serial_port = serial.Serial(port_name, timeout=1.0)  # 1 second timeout!

            serial_port.flushInput()  # deprecated function name;
            # use serial_port.reset_input_buffer()
            # if we can be sure that we have pySerial 3+.

            serial_port.write('v\r'.encode('ascii'))
            str_version = serial_port.readline()
            if str_version and str_version.startswith("EBB".encode('ascii')):
                return serial_port

            serial_port.write('v\r'.encode('ascii'))
            str_version = serial_port.readline()
            if str_version and str_version.startswith("EBB".encode('ascii')):
                return serial_port
            serial_port.close()
        except serial.SerialException as err:
            logger.error("Error testing serial port `{}` connection".format(port_name))
            logger.info("Error context:", exc_info=err)
        return None


def openPort():
    # Find and open a port to a single attached EiBotBoard.
    # The first port located will be used.
    found_port = findPort()
    serial_port = testPort(found_port)
    if serial_port:
        return serial_port


def open_named_port(port_name):
    # Find and open a port to a single attached EiBotBoard.
    # The first port located will be used.
    found_port = find_named_ebb(port_name)
    serial_port = testPort(found_port)
    if serial_port:
        return serial_port


def closePort(port_name):
    if port_name is not None:
        try:
            port_name.close()
        except serial.SerialException:
            pass


def query(port_name, cmd):
    if port_name is not None and cmd is not None:
        response = ''
        try:
            port_name.write(cmd.encode('ascii'))
            response = port_name.readline().decode('ascii')
            n_retry_count = 0
            while len(response) == 0 and n_retry_count < 100:
                # get new response to replace null response if necessary
                response = port_name.readline()
                n_retry_count += 1
            if cmd.split(",")[0].strip().lower() not in ["v", "i", "a", "mr", "pi", "qm"]:
                # Most queries return an "OK" after the data requested.
                # We skip this for those few queries that do not return an extra line.
                unused_response = port_name.readline()  # read in extra blank/OK line
                n_retry_count = 0
                while len(unused_response) == 0 and n_retry_count < 100:
                    # get new response to replace null response if necessary
                    unused_response = port_name.readline()
                    n_retry_count += 1
        except (serial.SerialException, IOError, RuntimeError, OSError) as err:
            logger.error("Error reading serial data")
            logger.info("Error context:", exc_info=err)
        return response


def command(port_name, cmd):
    if port_name is not None and cmd is not None:
        try:
            port_name.write(cmd.encode('ascii'))
            response = port_name.readline().decode('ascii')
            n_retry_count = 0
            while len(response) == 0 and n_retry_count < 100:
                # get new response to replace null response if necessary
                response = port_name.readline().decode('ascii')
                n_retry_count += 1
            if response.strip().startswith("OK"):
                # Debug option: indicate which command:
                # inkex.errormsg( 'OK after command: ' + cmd )
                pass
            else:
                if response:
                    error_msg = '\n'.join(('Unexpected response from EBB.',
                                           '    Command: {0}'.format(cmd.strip()),
                                           '    Response: {0}'.format(response.strip())))
                else:
                    error_msg = 'EBB Serial Timeout after command: {0}'.format(cmd)
                logger.error(error_msg)
        except (serial.SerialException, IOError, RuntimeError, OSError) as err:
            if cmd.strip().lower() not in ["rb"]: # Ignore error on reboot (RB) command
                logger.error('Failed after command: {0}'.format(cmd))
                logger.info("Error context:", exc_info=err)


def bootload(port_name):
    # Enter bootloader mode. Do not try to read back data.
    if port_name is not None:
        try:
            port_name.write('BL\r'.encode('ascii'))
            return True
        except:
            return False


def min_version(port_name, version_string):
    # Query the EBB firmware version for the EBB located at port_name.
    # Return True if the EBB firmware version is at least version_string.
    # Return False if the EBB firmware version is below version_string.
    # Return None if we are unable to determine True or False.

    if port_name is not None:
        ebb_version_string = queryVersion(port_name)  # Full string, human readable
        ebb_version_string = ebb_version_string.split("Firmware Version ", 1)

        if len(ebb_version_string) > 1:
            ebb_version_string = ebb_version_string[1]
        else:
            return None  # We haven't received a reasonable version number response.

        ebb_version_string = ebb_version_string.strip()  # Stripped copy, for number comparisons
        if LooseVersion(ebb_version_string) >= LooseVersion(version_string):
            return True
        else:
            return False


def queryVersion(port_name):
    return query(port_name, 'V\r')  # Query EBB Version String



# -*- coding: utf-8 -*-
# rtree.py
# part of plotink: https://github.com/evil-mad/plotink
#
# See below for version information
#
# Written by Michal Migurski https://github.com/migurski @michalmigurski
# as a contribution to the AxiDraw project https://github.com/evil-mad/axidraw/
#
# Copyright (c) 2022 Windell H. Oskay, Evil Mad Scientist Laboratories
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
rtree.py

Minimal R-tree spatial index class for calculating intersecting regions
"""

import math

def version():    # Version number for this document
    """Return version number of this script"""
    return "1.0" # Dated 2022-01-03

class Index:
    ''' One-shot R-Tree index (no rebalancing, insertions, etc.)
    '''
    bboxes = []
    subtrees = []
    xmin = None
    ymin = None
    xmax = None
    ymax = None

    def __init__(self, bboxes):
        center_x, center_y = 0, 0
        self.xmin, self.ymin = math.inf, math.inf
        self.xmax, self.ymax = -math.inf, -math.inf

        for (_, (xmin, ymin, xmax, ymax)) in bboxes:
            center_x += (xmin/2 + xmax/2) / len(bboxes)
            center_y += (ymin/2 + ymax/2) / len(bboxes)
            self.xmin = min(self.xmin, xmin)
            self.ymin = min(self.ymin, ymin)
            self.xmax = max(self.xmax, xmax)
            self.ymax = max(self.ymax, ymax)

        # Make four lists of bboxes, one for each quadrant around the center point
        # An original bbox may be present in more than one list
        sub_bboxes = [
            [
                (i, (x_1, y_1, x_2, y_2)) for (i, (x_1, y_1, x_2, y_2)) in bboxes
                if x_1 < center_x and y_1 < center_y
            ],
            [
                (i, (x_1, y_1, x_2, y_2)) for (i, (x_1, y_1, x_2, y_2)) in bboxes
                if x_2 > center_x and y_1 < center_y
            ],
            [
                (i, (x_1, y_1, x_2, y_2)) for (i, (x_1, y_1, x_2, y_2)) in bboxes
                if x_1 < center_x and y_2 > center_y
            ],
            [
                (i, (x_1, y_1, x_2, y_2)) for (i, (x_1, y_1, x_2, y_2)) in bboxes
                if x_2 > center_x and y_2 > center_y
            ],
        ]

        # Store bboxes or subtrees but not both
        if max(map(len, sub_bboxes)) == len(bboxes):
            # One of the subtrees is identical to the whole tree so just keep all the bboxes
            self.bboxes = bboxes
        else:
            # Make four subtrees, one for each quadrant
            self.subtrees = [Index(sub) for sub in sub_bboxes]

    def intersection(self, bbox):
        ''' Get a set of IDs for a given bounding box
        '''
        ids, (x_1, y_1, x_2, y_2) = set(), bbox

        for (i, (xmin, ymin, xmax, ymax)) in self.bboxes:
            is_disjoint = x_1 > xmax or y_1 > ymax or x_2 < xmin or y_2 < ymin
            if not is_disjoint:
                ids.add(i)

        for subt in self.subtrees:
            is_disjoint = x_1 > subt.xmax or y_1 > subt.ymax or x_2 < subt.xmin or y_2 < subt.ymin
            if not is_disjoint:
                ids |= subt.intersection(bbox)

        return ids



# coding=utf-8
# ebb_motion.py

"""
Motion control utilities for EiBotBoard
https://github.com/evil-mad/plotink

Intended to provide some common interfaces that can be used by
EggBot, WaterColorBot, AxiDraw, and similar machines.

See version() below for version number.

The MIT License (MIT)

Copyright (c) 2022 Windell H. Oskay, Evil Mad Scientist Laboratories

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import math
from . import ebb_serial

def version():  # Report version number for this document
    ''' Return version number '''
    return "0.22"  # Dated 2022-02-22


def doABMove(port_name, delta_a, delta_b, duration):
    '''
    Issue command to move A/B axes as: "XM,<move_duration>,<axisA>,<axisB><CR>"
    Then, <Axis1> moves by <AxisA> + <AxisB>, and <Axis2> as <AxisA> - <AxisB>
    '''
    if port_name is not None:
        str_output = 'XM,{0},{1},{2}\r'.format(duration, delta_a, delta_b)
        ebb_serial.command(port_name, str_output)


def doTimedPause(port_name, n_pause):
    ''' "Hardware" pause on EBB control board '''
    if port_name is not None:
        while n_pause > 0:
            if n_pause > 750:
                time_delay = 750
            else:
                time_delay = n_pause
                if time_delay < 1:
                    time_delay = 1  # don't allow zero-time moves
            ebb_serial.command(port_name, 'SM,{0},0,0\r'.format(time_delay))
            n_pause -= time_delay


def doLowLevelMove(port_name, rate1, steps1, accel1, rate2, steps2, accel2, clear=None):
    '''
    Execute a "pre-computed" 2D movement of the form
      "LM,<Rate1>,<Steps1>,<Accel1>,<Rate2>,<Steps2>,<Accel2>[,Clear]<CR>"
      See http://evil-mad.github.io/EggBot/ebb.html#LM for documentation.
    Requires firmware version 2.7.0 or higher for proper operation
    '''
    if port_name is not None:
        if ((rate1 == 0 and accel1 == 0) or steps1 == 0) and\
                ((rate2 == 0 and accel2 == 0) or steps2 == 0):
            return # No steps to take on either axis
        if clear:
            str_output = 'LM,{0},{1},{2},{3},{4},{5},{6}\r'.format(rate1,\
                                    steps1, accel1, rate2, steps2, accel2, clear)
        else:
            str_output = 'LM,{0},{1},{2},{3},{4},{5}\r'.format(rate1,\
                                steps1, accel1, rate2, steps2, accel2)
        ebb_serial.command(port_name, str_output)


def doXYMove(port_name, delta_x, delta_y, duration):
    '''
    Move X/Y axes as: "SM,<move_duration>,<axis1>,<axis2><CR>"
    Typically, this is wired up such that axis 1 is the Y axis and axis 2 is the X axis of motion.
    On EggBot, Axis 1 is the "pen" motor, and Axis 2 is the "egg" motor.
    '''
    if port_name is not None:
        str_output = 'SM,{0},{1},{2}\r'.format(duration, delta_y, delta_x)
        ebb_serial.command(port_name, str_output)


def doAbsMove(port_name, rate, position1=None, position2=None):
    '''
    Absolute XY or homing move as: "HM,<rate>[,<position1>,<position2>]<CR>"
        See http://evil-mad.github.io/EggBot/ebb.html#HM for more information

    This is an "absolute" move, to a position relative to where the motors
    were enabled. It is not necessarily a move in a straight line.
    If both position1 and position2 are given, then move to the given position.
    Otherwise, return to the position where the motors were enabled.
    
    This command requires firmware version 2.6.2 or newer for moves to
    the starting position, or firmware 2.7.0 or newer for moves that
    specify an XY position.
    
    This command DOES NOT perform a firmware version check. Use (e.g.)
        ebb_serial.min_version(port_name, "2.6.2") or
        ebb_serial.min_version(port_name, "2.7.0") if necessary.
    '''
    if port_name is not None:
        if position1 and position2:
            str_output = 'HM,{0},{1},{2}\r'.format(rate, position1, position2)
        else:
            str_output = 'HM,{0}\r'.format(rate)
        ebb_serial.command(port_name, str_output)


def moveDistLM(rate_in, accel_in, time_ticks):
    '''
    Calculate the number of motor steps taken using the LM command,
    with rate factor r, acceleration accel_in, and in a given number
    of 40 us time_ticks. Calculation is for one axis only.
    Step distance moved after T time ticks is given by:
        S = floor(( 2 * R * T + A * T^2 )/ 2^32 )

    This calculation is valid for the revised LM command as of
    version 2.7 of the EBB firmware.
    '''
    time = int(time_ticks)  # Ensure that the inputs are integral.
    rate_r0 = int(rate_in)
    accel = int(accel_in)
    if time == 0:
        return 0
    return (2 * rate_r0 * time + accel * time * time) >> 32


def moveDistLMA(rate_in, accel_in, time_ticks, accum_in):
    '''
    Calculate motor step count and final accumulator "remainder"
    after a certain number of time ticks, using the LM or LT
    command. Calculation is for one axis only.

    Inputs: Rate factor rate_in, acceleration accel_in, initial
    accumulator value accum_in, and 40 us intervals time_ticks.

    Accumulator value T time ticks is given by:
        Accum = R * T + 0.5 * accel_in * T^2 + accum_in
        Steps = floor (Accum / 2^31)
        Remainder = Accum - 2^31 * Steps

    Return Steps, Remainder
    This calculation is valid for version 2.7+ of the EBB firmware.
    '''
    time = int(time_ticks)  # Ensure that the inputs are integral.
    rate_r0 = int(rate_in)
    accel = int(accel_in)
    accum_0 = int(accum_in)
    if time == 0:
        return 0, 0

    # TWICE the accumulator value, before dividing:
    accum_end = (2 * accum_0 + 2 * rate_r0 * time + accel * time * time) >> 1
    step_count = accum_end >> 31
    rems = (accum_end) - (step_count << 31)

    return step_count, rems


def moveTimeLM(rate_in, steps, accel_in):
    """
    Calculate how long, in 40 us ISR intervals, the LM command will take to move one axis.

    First: Distance in steps moved after T time ticks is given by
      the formula: S = floor(( 2 * R * T + A * T^2 )/ 2^32 )
    Use the quadratic formula to solve for possible values of time T,
    the number of time ticks needed to travel the through distance of steps.

    As this is a floating point result, we will round down the output, and
    then move one time step forward until we find the result.

    This calculation is valid for the revised LM command as of
    version 2.7 of the EBB firmware.
    """

    steps = abs(steps)
    rate = float(rate_in)
    accel = float(accel_in)

    if steps == 0:
        return 0  # No steps to take, so takes zero time.

    if accel_in == 0:
        if rate_in == 0:
            return 0  # No move will be made if rate_in and accel_in are both zero.

        # Case of no acceleration: Simple to get actual movement time, since
        # Steps = floor(rate * Time / 2^31)
        # Thus, Time = ceil(Steps * 2^31 / rate)

        scaled_f = int(steps) << 31
        return int(math.ceil(scaled_f / rate))

    # Otherwise, accel_in is not zero.

    # Solve quadratic for T
    # S = floor(( 2 * R * T + A * T^2 )/ 2^32 )
    # -> (1/2)A T^2 + R T - 2^31 S = 0
    # a = A/2
    # b = R
    # c = -2^31 S
    #
    # T = (-b +/- sqrt(b^2 - 4 a c)) / 2 a

    # factors:
    a_fact = accel / 2.0
    b_fact = rate
    c_fact = -2147483648.0 * steps

    root_factor = b_fact * b_fact - 4 * a_fact * c_fact

    if root_factor < 0:
        root_factor = 0
    root = math.sqrt(root_factor)

    result1 = int(math.ceil((- b_fact + root) / accel))
    result2 = int(math.ceil((- b_fact - root) / accel))

    if result1 < 0 and result2 < 0:
        return -1  # No plausible roots; movement time must be positive

    if result1 < 0:
        time_ticks = result2  # Pick the positive root
    elif result2 < 0:
        time_ticks = result1  # Pick the positive root
    elif result2 < result1:  # If both are valid, pick the smaller value.
        time_ticks = result2
    else:
        time_ticks = result1
    return time_ticks


def QueryPenUp(port_name):
    """ Check if the pen is up, using QP. """
    if port_name is not None:
        pen_status = ebb_serial.query(port_name, 'QP\r')
        if pen_status[0] == '0':
            return False
        return True
    return False


def QueryPRGButton(port_name):
    """ Check if the button has been pressed, using QB. """
    if port_name is not None:
        return ebb_serial.query(port_name, 'QB\r')
    return None

def sendDisableMotors(port_name):
    """ Disable stepper motors with EM command """
    if port_name is not None:
        ebb_serial.command(port_name, 'EM,0,0\r')


def sendEnableMotors(port_name, res):
    """
    Enable both motors with EM command at selected resolution.
        If res == 0, -> Motor disabled
        If res == 1, -> 16X microstepping
        If res == 2, -> 8X microstepping
        If res == 3, -> 4X microstepping
        If res == 4, -> 2X microstepping
        If res == 5, -> No microstepping
    """
    if res < 0:
        res = 0
    if res > 5:
        res = 5
    if port_name is not None:
        ebb_serial.command(port_name, 'EM,{0},{0}\r'.format(res))

def query_enable_motors(port_name):
    """
    Read current state of motors and their resolution.
    Returns: res_1, res_2
        These are formatted the same way as the EM command:

        If res_x == 0, -> Motor disabled
        If res_x == 1, -> 16X microstepping
        If res_x == 2, -> 8X microstepping
        If res_x == 3, -> 4X microstepping
        If res_x == 4, -> 2X microstepping
        If res_x == 5, -> No microstepping

    This query uses PI ( http://evil-mad.github.io/EggBot/ebb.html#PI )
    to read the output state of the I/O pins that control the motor driver ICs
    """
    if port_name is not None:
        try:
            result = ebb_serial.query(port_name, 'PI,E,0\r') # Read motor 1 enable pin
            enable_1 = result.split("PI,")[1].strip() == "0"
            result = ebb_serial.query(port_name, 'PI,C,1\r') # Read motor 2 enable pin
            enable_2 = result.split("PI,")[1].strip() == "0"
            result = ebb_serial.query(port_name, 'PI,E,2\r') # Read MS1
            ms_1 = result.split("PI,")[1].strip() == "1"
            result = ebb_serial.query(port_name, 'PI,E,1\r') # Read MS2
            ms_2 = result.split("PI,")[1].strip() == "1"
            result = ebb_serial.query(port_name, 'PI,A,6\r') # Read MS3
            ms_3 = result.split("PI,")[1].strip() == "1"

            if ms_1 and ms_2 and ms_3:
                res_1 = 1 # 16X microstepping
            elif ms_1 and ms_2:
                res_1 = 2 # 8X microstepping
            elif ms_2:
                res_1 = 3 # 4X microstepping
            elif ms_1:
                res_1 = 4 # 2X microstepping
            else:
                res_1 = 5 # 2X microstepping

            res_2 = res_1
            if not enable_1:
                res_1 = 0
            if not enable_2:
                res_2 = 0

            return res_1, res_2
        except:
            return None, None
    return None, None


def sendPenDown(port_name, pen_delay, pin=None):
    """
    Lower pen with SP command
    Optionally, specify which pin to use
    """
    if port_name is not None:
        if pin:
            str_output = 'SP,0,{},{}\r'.format(pen_delay, pin)
        else:
            str_output = 'SP,0,{}\r'.format(pen_delay)
        ebb_serial.command(port_name, str_output)


def sendPenUp(port_name, pen_delay, pin=None):
    """
    Raise pen with SP command
    Optionally, specify which pin to use
    """
    if port_name is not None:
        if pin:
            str_output = 'SP,1,{},{}\r'.format(pen_delay, pin)
        else:
            str_output = 'SP,1,{0}\r'.format(pen_delay)
        ebb_serial.command(port_name, str_output)

def PBOutConfig(port_name, pin, state):
    """
    Enable an I/O pin. Pin: {0,1,2, or 3}. State: {0 or 1}.
    Note that B0 is used as an alternate pause button input.
    Note that B1 is used as the pen-lift servo motor output.
    Note that B3 is used as the EggBot engraver output.
    For use with a laser (or similar implement), pin 3 is recommended
    """
    if port_name is not None:
        # Set initial Bx pin value, high or low:
        str_output = 'PO,B,{0},{1}\r'.format(pin, state)
        ebb_serial.command(port_name, str_output)
        # Configure I/O pin Bx as an output
        str_output = 'PD,B,{0},0\r'.format(pin)
        ebb_serial.command(port_name, str_output)


def PBOutValue(port_name, pin, state):
    """
    Set state of the I/O pin. Pin: {0,1,2, or 3}. State: {0 or 1}.
    Set the pin as an output with OutputPinBConfigure before using this.
    """
    if port_name is not None:
        str_output = 'PO,B,{0},{1}\r'.format(pin, state)
        ebb_serial.command(port_name, str_output)


def TogglePen(port_name):
    """ Toggle pen state using TP """
    if port_name is not None:
        ebb_serial.command(port_name, 'TP\r')


def setPenDownPos(port_name, servo_max):
    """ Set pen down position using SC """
    if port_name is not None:
        ebb_serial.command(port_name, 'SC,5,{0}\r'.format(servo_max))
        # servo_max may be in the range 1 to 65535, in units of 83 ns intervals.
        # This sets the "Pen Down"position.
        # http://evil-mad.github.io/EggBot/ebb.html#SC


def setPenDownRate(port_name, pen_down_rate):
    """ Set pen lowering speed using SC """
    if port_name is not None:
        ebb_serial.command(port_name, 'SC,12,{0}\r'.format(pen_down_rate))
        # Set the rate of change of the servo when going down.
        # http://evil-mad.github.io/EggBot/ebb.html#SC


def setPenUpPos(port_name, servo_min):
    """ Set pen up position using SC """
    if port_name is not None:
        ebb_serial.command(port_name, 'SC,4,{0}\r'.format(servo_min))
        # servo_min may be in the range 1 to 65535, in units of 83 ns intervals.
        # This sets the "Pen Up" position.
        # http://evil-mad.github.io/EggBot/ebb.html#SC


def setPenUpRate(port_name, pen_up_rate):
    """ Set pen raising speed using SC """
    if port_name is not None:
        ebb_serial.command(port_name, 'SC,11,{0}\r'.format(pen_up_rate))
        # Set the rate of change of the servo when going up.
        # http://evil-mad.github.io/EggBot/ebb.html#SC


def setEBBLV(port_name, ebb_lv):
    """
    Set the EBB "Layer" Variable, an 8-bit number we can read and write.
    (Unrelated to document layers; name is an historical artifact.)
    """
    if port_name is not None:
        ebb_serial.command(port_name, 'SL,{0}\r'.format(ebb_lv))


def queryEBBLV(port_name):
    """
    Query the EBB "Layer" Variable, an 8-bit number we can read and write.
    (Unrelated to document layers; name is an historical artifact.)
    """
    if port_name is not None:
        value = ebb_serial.query(port_name, 'QL\r')
        try:
            ret_val = int(value)
            return ret_val
        except:
            return None
    return None


def queryVoltage(port_name):
    """ Query the EBB motor power supply input voltage. """
    if port_name is not None:
        if not ebb_serial.min_version(port_name, "2.2.3"):
            return True # Unable to read version, or version is below 2.2.3.
                        # In these cases, issue no voltage warning.
        raw_string = (ebb_serial.query(port_name, 'QC\r'))
        split_string = raw_string.split(",", 1)
        split_len = len(split_string)
        if split_len > 1:
            voltage_value = int(split_string[1])  # Pick second value only
        else:
            return True  # We haven't received a reasonable voltage string response.
            # Ignore voltage test and return.
        # Typical reading is about 300, for 9 V input.
        if voltage_value < 250:
            return False
    return True


def servo_timeout(port_name, timeout_ms, state=None):
    """
    Set the EBB servo motor timeout.
    The EBB will cut power to the pen-lift servo motor after a given
    time delay since the last X/Y/Z motion command.
    It can also optionally set an immediate on/off state.

    The time delay timeout_ms is given in ms. A value of 0 will
    disable the automatic power-off feature.

    The state parameter is given as 0 or 1, to turn off or on
    servo power immediately, respectively.

    This feature requires EBB hardware v 2.5.0 and firmware 2.6.0

    Reference: http://evil-mad.github.io/EggBot/ebb.html#SR
    """
    if port_name is not None:
        if not ebb_serial.min_version(port_name, "2.6.0"):
            return      # Unable to read version, or version is below 2.6.0.
        if state is None:
            str_output = 'SR,{0}\r'.format(timeout_ms)
        else:
            str_output = 'SR,{0},{1}\r'.format(timeout_ms, state)
        ebb_serial.command(port_name, str_output)



"""
Based on https://github.com/pypa/sampleproject

"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='plotink',
    version='1.4.2',
    python_requires='>=3.6.0',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/evil-mad/plotink',
    author='Evil Mad Scientist Laboratories',
    author_email='contact@evilmadscientist.com',
    description="Helper routines for use with plotters",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Intended Audience :: Developers",
    ],

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[
        'ink_extensions',
        'pyserial>=3.5',
    ],
    extras_require={
        'dev': ['coverage'],
        'test': [],
    },
)
